/* /////////////////////////////////////////////////////////////////////////////////////////////////
Benchmark 2021-10-21 by MacGruber
Plugin system for measuring VaM performance in a reliable way, making sure everyone uses the same settings.
https://www.patreon.com/MacGruber_Laboratory

Licensed under CC BY-SA. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;

namespace MacGruber
{
	public class Benchmark : MVRScript
	{
		private Atom myContainingAtom;
		private Transform myMNCanvas;
		private Transform myBMCanvas;
		private int myMode = MODE_INVALID;
		private int myASWMS = ASWMS_INVALID;
		private bool myHadWarnings = false;
		private bool myHadModifiedScene = false;


		private DateTime myBenchmarkTime;
		private VaMSettings myUsedSettings = new VaMSettings();
		private List<Record> mySceneRecords = new List<Record>();
		private FullRecord myCurrentRecord = new FullRecord();
		private Record myLiveRecord = new Record();
		private float myFadeClock = 0.0f;
		private float myRecordClock = 0.0f;
		private float myTextUpdateClock = 0.0f;
		private int myRecordClockSeconds = -1;
		private bool myEngaged = false;
		private bool myTextUpdateNeeded = false;

		private string myBundleURL = null;
		private string myVersion;
		private string myRunMode;
		private bool myIsInExpectedPackage = false;

		private static readonly string EXPECTED_PACKAGE_NAME = "MacGruber.Benchmark";
		private static readonly string ASSET_PATH = "Assets/Benchmark/";
		private static readonly string SAVE_DIRECTORY = "Saves/PluginData/Benchmark/";
		private static readonly string SUBSCENE_DIRECTORY = "Custom/SubScene/MacGruber/Benchmark/";
		private const bool LOG_SCENE_HASHES = false;
		private const bool LOG_SCENE_JSON = false;
		private static readonly string[] SECURED_FILES = new string[] {
			"Benchmark.json",
			"SelfCheck.json",
			"Warm-up.json",
			"Baseline1.json",
			"ClothSim.json",
			"Mirror.json",
			"Baseline2.json",
			"HairSim.json"
		};
		private static readonly string[] SECURED_HASHES = new string[] {
			"df075898ec7c24074a15fb465abf9fd7",
			"592922053e593982d35790609b1eb69f",
			"df7e246833600c2c203944ec4822ee70",
			"96e85d84d49ee5ce998d3b7a39519a66",
			"52cdcfffe20826fa0793d1d5d85223af",
			"e0b2ab48f8873139d3b315e0b53931d1",
			"9d503c6cdcf6322798bda8565b894fea",
			"e588082a08d3f26c5461d201a84371a9",
		};
		private static readonly string[] VAR_FILES = new string[] {
			"MacGruber.LogicBricks.10",
			"MacGruber.Life.13",
			"MacGruber.Essentials.11",
			"AcidBubbles.SpawnPoint.5",
			"AcidBubbles.Timeline.243",
			"AcidBubbles.Glance.16",
			"VL_13.Leggings_S.1",
			"VL_13.Top_SK.1",
			"prestigitis.clothing-200512.1",
			"NoStage3.Hair_Short_Pixie.1",
			"NoStage3.Hair_Long_Side_3.1"
		};

		private static readonly Color BASE_COLOR = new Color(0.0f, 0.0f, 0.0f, 0.8f);
		private static readonly Vector3 POSITION_OFFSET_DESKTOP = new Vector3(0.0f,0.8f,2.5f);
		private static readonly Quaternion ROTATION_OFFSET_DESKTOP = Quaternion.Euler(0.0f, 0.0f, 0.0f);
		private static readonly Vector3 SCALE_DESKTOP = Vector3.one * 0.004f;
		private static readonly Vector3 POSITION_OFFSET_VR = new Vector3(0.0f,0.4f,1.0f);
		private static readonly Quaternion ROTATION_OFFSET_VR = Quaternion.Euler(-21.801f, 0.0f, 0.0f);
		private static readonly Vector3 SCALE_VR = Vector3.one * 0.004f * 0.6f;
		private const float FADE_DURATION = 0.5f;
		private const float TEXT_DURATION = 0.5f;

		private const int MODE_INVALID = -1;
		private const int MODE_BENCHMARK_OFFICIAL = 0;
		private const int MODE_BENCHMARK_CUSTOM = 1;
		private const int MODE_BENCHMARK_FLUFF = 2;
		private const int MODE_ABORT = 3;

		private const int ASWMS_INVALID = -1;
		private const int ASWMS_ENABLED = 0;
		private const int ASWMS_DISABLED = 1;
		private const int ASWMS_UNKNOWN = 2;

		private JSONStorableAction myPlayScene;
		private JSONStorableString mySceneName;
		private JSONStorableAction myEndScene;
		private JSONStorableAction myEndBenchmark;

		private Material myUIMaterial;

		public override void Init()
		{
			myContainingAtom = containingAtom; // working around VaM weirdness, containingAtom is set to null before the plugin is fully destroyed.
			myBMCanvas = null;
			myBundleURL = Utils.GetPluginPath(this) + "/Internal/Benchmark.assetbundle";
			AssetLoader.AssetBundleFromFileRequest request = new AssetLoader.AssetBundleFromFileRequest {path = myBundleURL, callback = OnAssetBundleLoaded};
			AssetLoader.QueueLoadAssetBundleFromFile(request);

			string package = Utils.GetPackagePath(this);
			if (package != string.Empty)
			{
				package = package.Substring(0, package.Length-2);
			}
			if (package.StartsWith(EXPECTED_PACKAGE_NAME+"."))
			{
				// package prefix: "Creator.Name.Version:/", we want the "Version" part
				int end = package.Length-1;
				int start = package.LastIndexOf('.', end);
				myVersion = "Version " + package.Substring(start+1, end-start);
				myIsInExpectedPackage = true;
			}
			else
			{
				myVersion = "Custom Version";
			}

			myPlayScene = Utils.SetupAction(this, "PlayScene", null);

			mySceneName = new JSONStorableString("StartScene", "");
			mySceneName.isStorable = mySceneName.isRestorable = false;
			RegisterString(mySceneName);

			myEndScene = Utils.SetupAction(this, "EndScene", null);
			myEndBenchmark = Utils.SetupAction(this, "EndBenchmark", null);

			StartCoroutine(DoMenu());
		}

		private void OnAssetBundleLoaded(AssetLoader.AssetBundleFromFileRequest request)
		{
			GameObject mnPrefab = request.assetBundle?.LoadAsset<GameObject>(ASSET_PATH + "MainUI.prefab");
			if (mnPrefab != null)
			{
				Transform parent = myContainingAtom.reParentObject.Find("object/rescaleObject");
				myMNCanvas = Instantiate(mnPrefab, parent).transform;
				myMNCanvas.localEulerAngles = new Vector3(0, 180, 0);
				Canvas canvas = myMNCanvas.GetComponent<Canvas>();
				FixUICanvas(canvas, ref myUIMaterial);
				myContainingAtom.AddCanvas(canvas);
				myMNCanvas.gameObject.SetActive(false);
			}

			GameObject bmPrefab = request.assetBundle?.LoadAsset<GameObject>(ASSET_PATH + "BenchmarkUI.prefab");
			if (bmPrefab != null)
			{
				myBMCanvas = Instantiate(bmPrefab).transform;
				myBMCanvas.gameObject.SetActive(false);
			}
		}

		private void OnDestroy()
		{
			if (myMNCanvas != null)
			{
				myContainingAtom.RemoveCanvas(myMNCanvas.GetComponent<Canvas>());
				Destroy(myMNCanvas.gameObject);
			}
			if (myBMCanvas != null)
				Destroy(myBMCanvas.gameObject);
			if (myUIMaterial != null)
				Destroy(myUIMaterial);
			if (myBundleURL != null)
				AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
		}

		private IEnumerator DoMenu()
		{
			while (true)
			{
				// Wait for AssetBundle to load
				while (myMNCanvas == null || myBMCanvas == null)
					yield return null;

				{ // Launch
					Transform panel = myMNCanvas.Find("Launch");
					panel.Find("Launch/Header/Text").GetComponent<Text>().text = "Benchmark - " + myVersion;
					Button buttonOfficial = panel.Find("Launch/ButtonOfficial").GetComponent<Button>();
					buttonOfficial.onClick.RemoveAllListeners();
					buttonOfficial.onClick.AddListener(() => { myMode = MODE_BENCHMARK_OFFICIAL; });
					Button buttonCustom = panel.Find("Launch/ButtonCustom").GetComponent<Button>();
					buttonCustom.onClick.AddListener(() => { myMode = MODE_BENCHMARK_CUSTOM; });
					Button buttonFluff = panel.Find("Launch/ButtonFluff").GetComponent<Button>();
					buttonFluff.onClick.AddListener(() => { myMode = MODE_BENCHMARK_FLUFF; });
					Button buttonSettings = panel.Find("Launch/ButtonSettings").GetComponent<Button>();
					buttonSettings.onClick.AddListener(() => { VaMSettings.Benchmark.Apply(); });
					panel.gameObject.SetActive(true);
					myMNCanvas.gameObject.SetActive(true);

					// Wait for choosing mode
					while (myMode == MODE_INVALID)
						yield return null;

					myMNCanvas.gameObject.SetActive(false);
					panel.gameObject.SetActive(false);
					buttonOfficial.onClick.RemoveAllListeners();
					buttonCustom.onClick.RemoveAllListeners();
					buttonFluff.onClick.RemoveAllListeners();
					buttonSettings.onClick.RemoveAllListeners();
				}

				{ // Checks
					yield return DoChecks();

					if (myMode == MODE_INVALID)
						continue;
				}

				{ // Questions
					yield return DoQuestions();
				}

				{ // Benchmark
					VaMSettings restoreSettings = null;
					if (myMode == MODE_BENCHMARK_OFFICIAL)
					{
						restoreSettings = new VaMSettings();
						restoreSettings.Read();
						VaMSettings.Benchmark.Apply();
					}

					UserPreferences up = UserPreferences.singleton;
					HandModelControl hmc = up.motionHandModelControl;
					bool settingsAlwaysShowPointersOnTouch = up.alwaysShowPointersOnTouch;
					bool settingsLeftHandEnabled = hmc.leftHandEnabled;
					bool settingsRightHandEnabled = hmc.rightHandEnabled;
					up.alwaysShowPointersOnTouch = false;
					hmc.leftHandEnabled = false;
					hmc.rightHandEnabled = false;

					if (myMode == MODE_BENCHMARK_OFFICIAL)
						myRunMode = "Official Benchmark";
					else if (myMode == MODE_BENCHMARK_CUSTOM)
						myRunMode = "Custom Benchmark";
					else if (myMode == MODE_BENCHMARK_FLUFF)
						myRunMode = "Scene Only Run";
					else
						myRunMode = "Invalid Run";

					myBenchmarkTime = DateTime.UtcNow;
					myUsedSettings.Read();
					myMNCanvas.gameObject.SetActive(false);

					yield return DoBenchmark();

					up.alwaysShowPointersOnTouch = settingsAlwaysShowPointersOnTouch;
					hmc.leftHandEnabled = settingsLeftHandEnabled;
					hmc.rightHandEnabled = settingsRightHandEnabled;

					if (restoreSettings != null)
						restoreSettings.Apply();
				}

				if (myMode == MODE_INVALID)
				{
					SuperController.LogError("Benchmark: Something is wrong!");
					continue;
				}
				else if (myMode != MODE_ABORT) // Results
				{
					yield return DoDisplayResults();

					ResetData();
				}
				else // Abort
				{
					Transform panel = myMNCanvas.Find("Abort");
					Button buttonReset = panel.Find("ButtonReset").GetComponent<Button>();
					bool pressedReset = false;
					buttonReset.onClick.AddListener(() => { pressedReset = true; });

					panel.gameObject.SetActive(true);
					myMNCanvas.gameObject.SetActive(true);

					// Wait for pressing reset
					while (!pressedReset)
						yield return null;

					myMNCanvas.gameObject.SetActive(false);
					panel.gameObject.SetActive(false);
					buttonReset.onClick.RemoveAllListeners();
					ResetData();
				}
			}
		}

		private IEnumerator DoChecks()
		{
			myHadWarnings = false;
			myHadModifiedScene = false;
			Transform panel = myMNCanvas.Find("Messages");
			int previousChildCount = panel.childCount;

			CheckApplicationRuntime(panel);
			CheckSessionPlugins(panel);
			CheckDeclinedPlugins(panel);
			CheckIsInVARPackage(panel);
			CheckDependencies(panel);
			CheckSceneModifications(panel);

			if (panel.childCount == previousChildCount)
				yield break;

			myHadWarnings = true;
			int warningsCount = panel.childCount - previousChildCount;
			if (warningsCount > 3)
			{
				int minCount = 2;
				for (int i=previousChildCount+minCount; i<panel.childCount; ++i)
					panel.GetChild(i).gameObject.SetActive(false);

				CreateMessage(panel, true, 1,
					"<b>There are "+(warningsCount-minCount)+" additional warnings not shown. Fix above issues first!</b>"
				);
			}

			panel.Find("Header/Text").GetComponent<Text>().text = "Problems have been detected!";
			Transform buttons = panel.Find("Buttons");
			buttons.transform.SetAsLastSibling();
			Button buttonContinue = buttons.Find("ButtonContinue").GetComponent<Button>();
			Button buttonBack = buttons.Find("ButtonBack").GetComponent<Button>();

			bool pressedContinue = false;
			bool pressedBack = false;
			buttonContinue.onClick.AddListener(() => { pressedContinue = true; });
			buttonBack.onClick.AddListener(() => { pressedBack = true; });
			buttonContinue.gameObject.SetActive(myMode != MODE_BENCHMARK_OFFICIAL);
			buttonBack.gameObject.SetActive(true);

			panel.gameObject.SetActive(true);
			myMNCanvas.gameObject.SetActive(true);

			// Wait for pressing buttons
			while (!pressedContinue && !pressedBack)
				yield return null;

			myMNCanvas.gameObject.SetActive(false);
			panel.gameObject.SetActive(false);
			buttonContinue.onClick.RemoveAllListeners();
			buttonBack.onClick.RemoveAllListeners();

			for (int i=panel.childCount-1; i>=0; --i)
			{
				Transform child = panel.GetChild(i);
				if (child.name == "##TempEntry")
					Destroy(child.gameObject);
			}

			if (pressedBack)
				ResetData();
		}

		private void CreateMessage(Transform panel, bool isWarning, int lines, string message)
		{
			Transform prefab = panel.Find(isWarning ? "Warning" : "Message");
			RectTransform transform = Instantiate(prefab, panel).GetComponent<RectTransform>();
			transform.name = "##TempEntry";
			Vector2 sizeDelta = transform.sizeDelta;
			sizeDelta.y = 11 + Mathf.Max(0, lines-1) * 7;
			transform.sizeDelta = sizeDelta;
			transform.Find("Text").GetComponent<Text>().text = message;
			transform.gameObject.SetActive(true);
		}

		private void CheckSessionPlugins(Transform panel)
		{
			GameObject sessionPlugins = GameObject.Find("SceneAtoms/CoreControl/SessionPluginManagerContainer/CoreControl/SessionPluginManager/Plugins");
			int numSessionPlugins = sessionPlugins.transform.childCount;
			if (numSessionPlugins == 0)
				return;

			CreateMessage(panel, true, 10,
				"One or more SessionPlugins detected. These could potentially interfere with the benchmark. To make sure you have to temporarily remove them.\n\n" +
				"<b>Please do the following:</b>\n" +
				"1. Save your current session plugin setup as a preset.\n" +
				"2. Remove all session plugins and save that as default.\n" +
				"3. Restart VaM.\n" +
				"4. Reload this scene.\n" +
				"5. Later, after you are done with the benchmark, you can restore your setup from the\n" +
				"    preset and set it as default again."
			);
		}

		private void CheckDeclinedPlugins(Transform panel)
		{
			bool notFound = true;
			Color pluginPanelWarningColor = new Color(1f, 0.95f, 0.5f);
			MVRPluginUI[] plugins = Resources.FindObjectsOfTypeAll<MVRPluginUI>();
			for (int i=0; i<plugins.Length && notFound; ++i)
			{
				Image image = plugins[i].GetComponent<Image>();
				if (image != null && image.color == pluginPanelWarningColor)
					notFound = false;
			}

			if (notFound)
				return;

			CreateMessage(panel, true, 11,
				"One or more plugins in this scene had their security prompt denied by the user. You should only accept plugins from sources you trust. However, this benchmark can not function properly without having all its plugins.\n\n" +
				"<b>Please try the following:</b>\n" +
				"1. Reload this scene.\n" +
				"2. Accept all security prompts for this scene.\n\n" +
				"If that doesn't help, try deleting the content of your <i>AddonPackagesUserPrefs</i> folder and\n" +
				"restarting VaM. This will make VaM ask again for any plugins you may have set to\n" +
				"<i>Deny Always</i> in the past."
			);
		}

		private void CheckSceneModifications(Transform panel)
		{
			bool valid = true;

			JSONClass json = SuperController.singleton.GetSaveJSON();

			// remove stuff we need to ignore
			json.Remove("headPossessedController");
			json.Remove("playerHeightAdjust");
			json.Remove("monitorCameraRotation");
			JSONArray atoms = json["atoms"].AsArray;
			for (int i=0; i<atoms.Count; ++i)
			{
				string id = atoms[i].AsObject["id"];
				if (id == "[CameraRig]" || id == "Benchmark/Person")
				{
					atoms.Remove(i);
					--i;
				}
			}

			// deleted too much...we need to re-add some things :D
			JSONClass personPlugins = GetAtomById("Benchmark/Person")?.GetStorableByID("PluginManager")?.GetJSON();
			if (personPlugins != null)
				json.Add("Benchmark/Person@PluginManager", personPlugins);


			StringBuilder b = new StringBuilder();
			json.ToString("", b);
			string data = b.ToString();

			// working around float inaccuracy
			Regex rgx = new Regex("\"(-?\\d+\\.\\d{4})\\d+((?:E[\\+-]\\d+)?)\"");
			data = rgx.Replace(data, "\"$1$2\"");

			byte[] bytes = Encoding.ASCII.GetBytes(data);
			string hash = manager.GetMD5Hash(bytes);
			valid &= (hash == SECURED_HASHES[0]);
			if (LOG_SCENE_HASHES)
			{
				SuperController.LogMessage("Hashes:");
				SuperController.LogMessage("\"" + hash + "\",");
			}
			if (LOG_SCENE_JSON)
			{
				string sceneFile = ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString() + ".json";
				FileManagerSecure.CreateDirectory(SAVE_DIRECTORY);
				FileManagerSecure.WriteAllBytes(SAVE_DIRECTORY + sceneFile, bytes);
			}

			string subScenesPath = Utils.GetPackagePath(this) + SUBSCENE_DIRECTORY;
			for (int i=1; i<SECURED_FILES.Length; ++i)
			{
				string filename = subScenesPath + SECURED_FILES[i];
				try {
					bytes = FileManagerSecure.ReadAllBytes(filename);
				}
				catch (Exception e)
				{
					SuperController.LogError("Benchmark subscene: \"" + filename + "\" not found!");
					valid = false;
				}
				hash = manager.GetMD5Hash(bytes);
				valid &= (hash == SECURED_HASHES[i]);
				if (LOG_SCENE_HASHES)
					SuperController.LogMessage("\"" + hash + "\",");
			}

			if (valid || myHadModifiedScene)
				return;

			myHadModifiedScene = true;
			CreateMessage(panel, true, 10,
				"It appears you are using a modified version of the Benchmark. For an official run it is important that everyone uses <i>exactly</i> the same scene to have comparable results.\n\n" +
				"<b>Please do the following:</b>\n" +
				"1. Open the <i>Hub-Browse</i> window within VaM, search for <i>'Benchmark'</i> by <i>'MacGruber'</i> and\n" +
				"    use the <i>'Download All'</i> button. This should download all needed dependencies as well.\n" +
				"    Its <i>strongly</i> recommend <i>not</i> to retrieve it from any other source.\n" +
				"2. When reloading the scene, make sure you load directly from the\n" +
				"    <i>"+EXPECTED_PACKAGE_NAME+"</i> VAR package.\n" +
				"3. Do not unpack the VAR, do not modify anything!"
			);
		}

		private void CheckDependencies(Transform panel)
		{
			bool valid = true;
			for (int i=0; i<VAR_FILES.Length; ++i)
				valid &= FileManagerSecure.FileExists(VAR_FILES[i] + ":/meta.json");

			if (valid || myHadModifiedScene)
				return;

			myHadModifiedScene = true;
			CreateMessage(panel, true, 11,
				"One or more VAR dependencies of this scene could not be found. For an official run it is important that everyone uses <i>exactly</i> the same scene to have comparable results.\n\n" +
				"<b>Please do the following:</b>\n" +
				"1. Open the <i>Hub-Browse</i> window within VaM, search for <i>'Benchmark'</i> by <i>'MacGruber'</i> and\n" +
				"    use the <i>'Download All'</i> button. This should download all needed dependencies as well.\n" +
				"    Its <i>strongly</i> recommend <i>not</i> to retrieve it from any other source.\n" +
				"2. When reloading the scene, make sure you load directly from the\n" +
				"    <i>"+EXPECTED_PACKAGE_NAME+"</i> VAR package.\n" +
				"3. Do not unpack the VAR, do not modify anything!"
			);
		}

		private void CheckIsInVARPackage(Transform panel)
		{
			if (myIsInExpectedPackage || myHadModifiedScene)
				return;

			myHadModifiedScene = true;
			CreateMessage(panel, true, 11,
				"It seems the scene is not running from the <i>"+EXPECTED_PACKAGE_NAME+"</i> VAR package. For an official run it is important that everyone uses <i>exactly</i> the same scene to have comparable results.\n\n" +
				"<b>Please do the following:</b>\n" +
				"1. Open the <i>Hub-Browse</i> window within VaM, search for <i>'Benchmark'</i> by <i>'MacGruber'</i> and\n" +
				"    use the <i>'Download All'</i> button. This should download all needed dependencies as well.\n" +
				"    Its <i>strongly</i> recommend <i>not</i> to retrieve it from any other source.\n" +
				"2. When reloading the scene, make sure you load directly from the\n" +
				"    <i>"+EXPECTED_PACKAGE_NAME+"</i> VAR package.\n" +
				"3. Do not unpack the VAR, do not modify anything!"
			);
		}

		private void CheckApplicationRuntime(Transform panel)
		{
			// Float values have 24bit precision, roughly 7 decimal digits.
			// Hence we set a limit at half an hour worth of milliseconds.
			// Pro-Tip: Don't trust Unity's time measurements ;)

			float maxRuntimeMS = 30*60*1000; // 30min
			float elapsedMS = GlobalStopwatch.GetElapsedMilliseconds();
			if (elapsedMS <= maxRuntimeMS)
				return;

			CreateMessage(panel, true, 6,
				"VaM has been running for more than 30 minutes. This makes VaM's time measurement lose precision, leading to useless results.\n\n" +
				"<b>Please do the following:</b>\n" +
				"1. Restart VaM\n" +
				"2. Reload this scene."
			);
		}

		private IEnumerator DoQuestions()
		{
			SuperController sc = SuperController.singleton;

			if (sc.isOVR || sc.isOpenVR)
			{
				Transform panel = myMNCanvas.Find("Questions");
				Transform buttons = panel.Find("Buttons");
				buttons.transform.SetAsLastSibling();
				Button buttonEnabled = buttons.Find("ButtonEnabled").GetComponent<Button>();
				Button buttonDisabled = buttons.Find("ButtonDisabled").GetComponent<Button>();
				Button buttonMaybe = panel.Find("ButtonMaybe").GetComponent<Button>();

				myASWMS = ASWMS_INVALID;
				buttonEnabled.onClick.AddListener(() => { myASWMS = ASWMS_ENABLED; });
				buttonDisabled.onClick.AddListener(() => { myASWMS = ASWMS_DISABLED; });
				buttonMaybe.onClick.AddListener(() => { myASWMS = ASWMS_UNKNOWN; });
				panel.gameObject.SetActive(true);
				myMNCanvas.gameObject.SetActive(true);

				// Wait for pressing buttons
				while (myASWMS == ASWMS_INVALID)
					yield return null;

				myMNCanvas.gameObject.SetActive(false);
				panel.gameObject.SetActive(false);
				buttonEnabled.onClick.RemoveAllListeners();
				buttonDisabled.onClick.RemoveAllListeners();
				buttonMaybe.onClick.RemoveAllListeners();
			}
		}

		private IEnumerator DoBenchmark()
		{
			SuperController sc = SuperController.singleton;

			// setup
			Image panelScene = myBMCanvas.Find("Benchmark").GetComponent<Image>();
			Image panelTime = myBMCanvas.Find("Time").GetComponent<Image>();
			Image panelFpsAvg = myBMCanvas.Find("AVG").GetComponent<Image>();
			Image panelFpsMin = myBMCanvas.Find("MIN").GetComponent<Image>();
			Image panelFpsMax = myBMCanvas.Find("MAX").GetComponent<Image>();
			Text textScene = panelScene.transform.Find("Value").GetComponent<Text>();
			Text textTime = panelTime.transform.Find("Value").GetComponent<Text>();
			Text textFpsAvg = panelFpsAvg.transform.Find("Value").GetComponent<Text>();
			Text textFpsMin = panelFpsMin.transform.Find("Value").GetComponent<Text>();
			Text textFpsMax = panelFpsMax.transform.Find("Value").GetComponent<Text>();
			panelScene.transform.Find("Label").GetComponent<Text>().text = myRunMode;

			bool isVR = sc.isOVR || sc.isOpenVR;
			if (isVR == sc.MonitorRig.gameObject.activeSelf)
				sc.ToggleMainMonitor();

			bool previousMonitorUIActive = sc.MonitorUI.gameObject.activeSelf;
			SuperController.GameMode previousGameMode = sc.gameMode;
			sc.HideMainHUD();
			sc.HideMonitorUI();
			sc.gameMode = SuperController.GameMode.Play;


			bool endBenchmark = false;
			myPlayScene.actionCallback = PlaySceneAction;
			mySceneName.valNoCallback = string.Empty;
			mySceneName.setCallbackFunction = StartSceneAction;
			myEndScene.actionCallback = EndSceneAction;
			myEndBenchmark.actionCallback = () => {
				EndSceneAction();
				endBenchmark = true;
			};

			// start benchmark
			JSONStorableBool disableNavigation = myContainingAtom.GetStorableByID("plugin#0_MacGruber.CameraLock")?.GetBoolJSONParam("DisableNavigation");
			JSONStorableString switchState = myContainingAtom.GetStorableByID("plugin#1_MacGruber.StateMachine")?.GetStringJSONParam("SwitchState");
			if (disableNavigation == null)
				SuperController.LogMessage("Can't start benchmark, CameraLock brick not found!");
			if (switchState == null)
				SuperController.LogMessage("Can't start benchmark, StateMachine brick not found!");
			if (disableNavigation == null || switchState == null)
				yield break;
				
			disableNavigation.val = (myMode != MODE_BENCHMARK_FLUFF);
			switchState.val =  "Benchmark#Start";

			// run benchmark
			while (!endBenchmark)
			{
				if (myMode != MODE_ABORT && HaveAbortInput())
				{
					myMode = MODE_ABORT;
					EndSceneAction();
					switchState.val =  "Benchmark#Abort";
				}

				bool currentRecordResetNeeded = false;
				if (myEngaged)
				{
					myRecordClock += Time.deltaTime;
					myTextUpdateClock += Time.deltaTime;

					myCurrentRecord.Sample();
					myLiveRecord.Sample();

					if (myTextUpdateClock > TEXT_DURATION)
					{
						myLiveRecord.Apply();
						myTextUpdateClock -= TEXT_DURATION;
						myTextUpdateNeeded = true;
						currentRecordResetNeeded = true;
					}
				}

				float delta = Time.deltaTime / FADE_DURATION;
				myFadeClock += myEngaged ? delta : -delta;
				myFadeClock = Mathf.Clamp01(myFadeClock);
				if (myFadeClock > 0.0f)
				{
					Color c = BASE_COLOR;
					c.a *= Mathf.SmoothStep(0.0f, 1.0f, myFadeClock);
					panelScene.color = c;
					panelTime.color = c;
					panelFpsAvg.color = c;
					panelFpsMin.color = c;
					panelFpsMax.color = c;

					Transform cameraControl = CameraTarget.centerTarget.transform;
					Vector3 position = isVR ? POSITION_OFFSET_VR : POSITION_OFFSET_DESKTOP;
					Quaternion rotation = isVR ? ROTATION_OFFSET_VR : ROTATION_OFFSET_DESKTOP;
					myBMCanvas.localScale = isVR ? SCALE_VR : SCALE_DESKTOP;
					myBMCanvas.position = cameraControl.TransformPoint(position);
					myBMCanvas.rotation = cameraControl.rotation * rotation;
				}
				if (myTextUpdateNeeded)
				{
					if (myLiveRecord.Valid)
					{
						textFpsAvg.text = Mathf.RoundToInt(myLiveRecord.FPS.Avg) + "fps";
						textFpsMin.text = Mathf.RoundToInt(myLiveRecord.FPS.Min) + "fps";
						textFpsMax.text = Mathf.RoundToInt(myLiveRecord.FPS.Max) + "fps";
					}
					else
					{
						textFpsAvg.text = "---";
						textFpsMin.text = "---";
						textFpsMax.text = "---";
					}

					if (textScene.text != myLiveRecord.Scene)
						textScene.text = myLiveRecord.Scene;

					int recordClockSeconds = Mathf.RoundToInt(myRecordClock);
					if (recordClockSeconds != myRecordClockSeconds)
					{
						myRecordClockSeconds = recordClockSeconds;
						int seconds = recordClockSeconds % 60;
						int minutes = recordClockSeconds / 60;
						textTime.text = string.Format("{0:00}:{1:00}", minutes, seconds);
					}

					myTextUpdateNeeded = false;
				}

				myBMCanvas.gameObject.SetActive(myFadeClock > 0.0f);

				if (currentRecordResetNeeded)
				{
					myLiveRecord.Reset();
				}

				yield return null;
			}

			// cleanup
			myBMCanvas.gameObject.SetActive(false);
			myFadeClock = 0.0f;
			myPlayScene.actionCallback = null;
			mySceneName.setCallbackFunction = null;
			myEndScene.actionCallback = null;
			myEndBenchmark.actionCallback = null;

			if (isVR == sc.MonitorRig.gameObject.activeSelf)
				sc.ToggleMainMonitor();

			sc.gameMode = previousGameMode;
			if (previousMonitorUIActive)
				sc.ShowMonitorUI();
		}

		private bool HaveAbortInput()
		{
			SuperController sc = SuperController.singleton;
			bool isVR = sc.isOVR || sc.isOpenVR;
			return sc.GetLeftSelect()
				|| sc.GetMenuMoveLeft()
				|| sc.GetRightSelect()
				|| sc.GetMenuMoveRight()
				|| (sc.isOVR && !OVRManager.hasInputFocus)
				|| (isVR == sc.MonitorRig.gameObject.activeSelf)
				|| sc.mainHUD.gameObject.activeSelf
				|| sc.MonitorUI.gameObject.activeSelf
				|| Input.GetKey(KeyCode.Escape);
		}

		private void PlaySceneAction()
		{
			Atom atom = GetAtomById("Benchmark/CameraTarget");
			JSONStorable storable = atom?.GetStorableByID("plugin#0_VamTimeline.AtomPlugin");
			JSONStorableAction action = storable?.GetAction("Play");
			if (action == null)
				SuperController.LogError("Can't find 'Play' action in Benchmark subscene. Something is broken!");
			else
				action.actionCallback();
		}

		private void StartSceneAction(string sceneName)
		{
			mySceneName.valNoCallback = string.Empty;

			if (myEngaged)
				EndSceneAction();

			myLiveRecord.Init(sceneName);
			myTextUpdateClock = 0.0f;
			myTextUpdateNeeded = true;
			myRecordClock = 0.0f;
			myRecordClockSeconds = -1;

			myCurrentRecord.Init(sceneName);
			myEngaged = true;
		}

		private void EndSceneAction()
		{
			if (!myEngaged)
				return;
			myLiveRecord.Apply();
			myTextUpdateNeeded = true;

			if (!myCurrentRecord.Scene.StartsWith("Warm-up"))
				mySceneRecords.Add(myCurrentRecord.ApplyAndGetRecord());
			myCurrentRecord.Reset();

			myEngaged = false;
		}

		private void ResetData()
		{
			EndSceneAction();
			mySceneRecords.Clear();
			myCurrentRecord.Reset();
			myLiveRecord.Reset();
			myTextUpdateNeeded = true;
			myRecordClock = 0.0f;
			myRecordClockSeconds = -1;
			myTextUpdateClock = 0.0f;
			myMode = MODE_INVALID;
		}

		private IEnumerator DoDisplayResults()
		{
			SuperController sc = SuperController.singleton;
			UserPreferences up = UserPreferences.singleton;
			Transform panel = myMNCanvas.Find("Results");
			StringBuilder b = new StringBuilder();

			{ // Benchmark Info
				string header = myRunMode + " - " + myVersion;
				if (myHadWarnings)
					header += " - <color=red>Ignored warnings!</color>";

				panel.Find("Header/Text").GetComponent<Text>().text = header;

				b.Length = 0;
				b.Append("VaM: ").AppendLine(sc.GetVersion());
				b.Append("Name: ").AppendLine(up.creatorName);
				b.AppendFormat("{0:yyyy-MM-dd HH:mm} UTC", myBenchmarkTime).AppendLine();
				panel.Find("Overview/BenchmarkInfo/Text").GetComponent<Text>().text = b.ToString();
			}

			{ // Overall Result
				Transform overall = panel.Find("Overview/OverallResult");
				Color color = new Color(1.0f, 0.0f, 0.0f, 200.0f/255.0f);
				if (myMode == MODE_BENCHMARK_OFFICIAL)
					color = new Color(0.0f, 0.75f, 0.0f, 200.0f/255.0f);
				overall.GetComponent<Image>().color = color;

				float avg = 0.0f;
				float min = float.MaxValue;
				float max = 0.0f;
				for (int i=0; i<mySceneRecords.Count; ++i)
				{
					Record r = mySceneRecords[i];
					avg += r.FPS.Avg;
					min = Mathf.Min(min, r.FPS.Min);
					max = Mathf.Max(max, r.FPS.Max);
				}
				if (mySceneRecords.Count > 0)
				{
					avg /= mySceneRecords.Count;
				}

				Transform grid = overall.Find("Grid");
				grid.Find("FPS.AVG").GetComponent<Text>().text = string.Format("{0:F2}", avg);
				grid.Find("FPS.MIN").GetComponent<Text>().text = string.Format("{0:F2}", min);
				grid.Find("FPS.MAX").GetComponent<Text>().text = string.Format("{0:F2}", max);
			}

			{ // System Info
				b.Length = 0;
				float graphicsMemorySize = RoundOneDigit(SystemInfo.graphicsMemorySize / 1024.0f);
				float systemMemorySize = RoundOneDigit(SystemInfo.systemMemorySize / 1024.0f);
				b.AppendLine("<color=white><b>System Information</b></color>");
				if (sc.isOVR || sc.isOpenVR)
					b.Append("Headset Model: ").AppendLine(XRDevice.model);
				b.Append("CPU: ").AppendLine(SystemInfo.processorType);
				b.Append("GPU: ").Append(SystemInfo.graphicsDeviceName).AppendFormat(" (approx. {0} GB)", graphicsMemorySize).AppendLine();
				b.Append("MBO: ").AppendLine(SystemInfo.deviceModel);
				b.Append("RAM: ").AppendFormat("{0} GB", systemMemorySize).AppendLine();
				b.Append("OS:  ").AppendLine(SystemInfo.operatingSystem);
				b.AppendLine();
				b.AppendLine("Note that this basic info does NOT detect any overclocking of your system. You need that kind of info and other relevant facts to the saved data by hand.");
				panel.Find("Overview/SystemInfo/Text").GetComponent<Text>().text = b.ToString();
			}


			{ // VaM Settings
				VaMSettings official = VaMSettings.Benchmark;

				b.Length = 0;
				b.AppendLine("<color=white><b>VaM Settings</b></color>");
				b.Append("<color=red>Platform: ");
				if (sc.isOVR)
					b.Append("OculusVR");
				else if (sc.isOpenVR)
					b.Append("OpenVR");
				else
					b.Append("Desktop");
				b.AppendLine("</color>");

				if (sc.isOVR || sc.isOpenVR)
				{
					b.Append("ASW/MS: ");
					if (myASWMS == ASWMS_ENABLED)
						b.AppendLine("<color=red>On !!!</color>");
					else if (myASWMS == ASWMS_DISABLED)
						b.AppendLine("Off");
					else if (myASWMS == ASWMS_UNKNOWN)
						b.AppendLine("<color=red>Unknown !!!</color>");
					else // ASWMS_INVALID
						b.AppendLine("<color=red>Error !!!</color>");

					float refreshRate = RoundOneDigit(XRDevice.refreshRate);
					b.AppendFormat("<color=red>Internal Resolution: {0}x{1}</color>", 2*XRSettings.eyeTextureWidth, XRSettings.eyeTextureHeight).AppendLine();
					b.AppendFormat("<color=red>Resolution Scale: {0}x</color>", XRSettings.eyeTextureResolutionScale).AppendLine();
					b.AppendFormat("Refresh Rate: {0}Hz", refreshRate).AppendLine();
					b.AppendLine(Compare("Mirror To Desktop: ", official.mirrorToDisplay, myUsedSettings.mirrorToDisplay));
				}
				else
				{
					float refreshRate = RoundOneDigit(Screen.currentResolution.refreshRate);
					b.AppendFormat("<color=red>Resolution: {0}x{1} @ {2}Hz", Screen.width, Screen.height, refreshRate).AppendLine();
					b.Append("Mode: ").Append(Screen.fullScreenMode.ToString()).AppendLine("</color>");
					b.AppendLine(Compare("Desktop VSync: ", official.desktopVsync, myUsedSettings.desktopVsync));
				}

				b.AppendLine(Compare("Time Scale: ", official.timeScale, myUsedSettings.timeScale, ""));

				if (!sc.isOVR && !sc.isOpenVR)
					b.AppendLine(Compare("Field of View: ", official.monitorCameraFOV, myUsedSettings.monitorCameraFOV, "°"));

				b.AppendLine(Compare("Mirrored Surfaces: ", official.mirrorReflections, myUsedSettings.mirrorReflections));
				b.AppendLine(Compare("Realtime ReflectionProbes: ", official.realtimeReflectionProbes, myUsedSettings.realtimeReflectionProbes));
				b.AppendLine(Compare("MSAA Level: ", MSAAToStr(official.msaaLevel), MSAAToStr(myUsedSettings.msaaLevel)));
				b.AppendLine(Compare("Pixel Light Count: ", official.pixelLightCount, myUsedSettings.pixelLightCount, ""));
				b.AppendLine(Compare("Shader Quality: ", official.shaderLOD.ToString(), myUsedSettings.shaderLOD.ToString()));
				b.AppendLine(Compare("Glow Effects: ", official.glowEffects.ToString(), myUsedSettings.glowEffects.ToString()));
				b.AppendLine(Compare("Smooth Passes: ", official.smoothPasses, myUsedSettings.smoothPasses, ""));
				b.AppendLine(Compare("Physics Rate: ", PhysicsToStr(official.physicsRate), PhysicsToStr(myUsedSettings.physicsRate)));
				b.AppendLine(Compare("Physics Update Cap: ", official.physicsUpdateCap, myUsedSettings.physicsUpdateCap, ""));
				b.AppendLine(Compare("Soft Body Physics: ", official.softPhysics, myUsedSettings.softPhysics));
				b.AppendLine(Compare("High Quality Physics: ", official.physicsHighQuality, myUsedSettings.physicsHighQuality));

				panel.Find("Settings/Settings/Text").GetComponent<Text>().text = b.ToString();
			}

			{ // Results
				int count = Mathf.Min(mySceneRecords.Count, 6);
				for (int i=0; i<count; ++i)
				{
					Transform resultPanel = i < 3 ? panel.Find("ResultsLeft/Result"+i) : panel.Find("ResultsRight/Result"+i);
					resultPanel.Find("Description/Scene").GetComponent<Text>().text = mySceneRecords[i].Scene;
					Transform grid = resultPanel.Find("Grid");

					Record.MinMaxAvg data;
					data = mySceneRecords[i].FPS;
					grid.Find("FPS.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("FPS.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("FPS.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					data = mySceneRecords[i].TotalTime;
					grid.Find("TotalTime.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("TotalTime.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("TotalTime.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					data = mySceneRecords[i].RenderTime;
					grid.Find("RenderTime.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("RenderTime.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("RenderTime.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					data = mySceneRecords[i].ScriptsTime;
					grid.Find("ScriptsTime.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("ScriptsTime.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("ScriptsTime.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					data = mySceneRecords[i].PhysicsTime;
					grid.Find("PhysicsTime.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("PhysicsTime.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("PhysicsTime.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					data = mySceneRecords[i].WaitTime;
					grid.Find("WaitTime.AVG").GetComponent<Text>().text = string.Format("{0:F2}", data.Avg);
					grid.Find("WaitTime.MIN").GetComponent<Text>().text = string.Format("{0:F2}", data.Min);
					grid.Find("WaitTime.MAX").GetComponent<Text>().text = string.Format("{0:F2}", data.Max);

					for (int c=0; c<resultPanel.childCount; ++c)
						resultPanel.GetChild(c).gameObject.SetActive(true);
				}
				for (int i=count; i<6; ++i)
				{
					Transform resultPanel = i < 3 ? panel.Find("ResultsLeft/Result"+i) : panel.Find("ResultsRight/Result"+i);
					for (int c=0; c<resultPanel.childCount; ++c)
						resultPanel.GetChild(c).gameObject.SetActive(false);
				}
			}

			Button buttonSave = panel.Find("ButtonSave").GetComponent<Button>();
			Button buttonReset = panel.Find("ButtonReset").GetComponent<Button>();

			buttonSave.interactable = true;
			buttonSave.onClick.AddListener(() => {
				buttonSave.interactable = false;
				SaveResults();
			});

			bool pressedReset = false;
			buttonReset.onClick.AddListener(() => { pressedReset = true; });

			panel.gameObject.SetActive(true);
			myMNCanvas.gameObject.SetActive(true);

			// Wait for pressing reset
			while (!pressedReset)
				yield return null;

			myMNCanvas.gameObject.SetActive(false);
			panel.gameObject.SetActive(false);

			buttonSave.onClick.RemoveAllListeners();
			buttonReset.onClick.RemoveAllListeners();
		}

		private void SaveResults()
		{
			// prepare
			int width = 3000;
			int height = 1030;
			Transform panel = myMNCanvas.Find("Results");
			Camera camera = panel.Find("ScreenshotCamera")?.GetComponent<Camera>();
			GameObject buttonSave = panel.Find("ButtonSave").gameObject;
			GameObject buttonReset = panel.Find("ButtonReset").gameObject;

			int previousLayer = panel.gameObject.layer;
			bool previousActive = panel.gameObject.activeSelf;
			int screenshotLayer = LayerMask.NameToLayer("Water"); // hackydy hack
			myMNCanvas.gameObject.layer = screenshotLayer;
			panel.gameObject.SetActive(true);
			buttonSave.SetActive(false);
			buttonReset.SetActive(false);

			// render
			RenderTexture rt = RenderTexture.GetTemporary(width, height, 24, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB, 8);
			camera.targetTexture = rt;
			camera.cullingMask = 1 << screenshotLayer;
			camera.Render();

			// save image
			Texture2D texture = new Texture2D(width, height, TextureFormat.RGB24, false);
			RenderTexture previousRT = RenderTexture.active;
			RenderTexture.active = rt;
			texture.ReadPixels(new Rect(0, 0, width, height), 0, 0, false);
			texture.Apply();
			string filename = SAVE_DIRECTORY + "Benchmark-" + myBenchmarkTime.ToString("yyyyMMdd-HHmmss") + ".png";
			FileManagerSecure.CreateDirectory(SAVE_DIRECTORY);
			byte[] bytes = texture.EncodeToPNG();
			FileManagerSecure.WriteAllBytes(filename, bytes);

			// cleanup
			RenderTexture.active = previousRT;
			RenderTexture.ReleaseTemporary(rt);
			Destroy(texture);
			panel.gameObject.SetActive(previousActive);
			buttonSave.SetActive(true);
			buttonReset.SetActive(true);
			myMNCanvas.gameObject.layer = previousLayer;
		}

		private static float Round(float x)
		{
			return (int)(x + 0.5f);
		}

		private static float RoundOneDigit(float x)
		{
			return ((int)(x*10.0f + 0.5f)) / 10.0f;
		}

		private static string PhysicsToStr(UserPreferences.PhysicsRate physicsRate)
		{
			switch (physicsRate)
			{
				case UserPreferences.PhysicsRate.Auto: return "Auto";
				case UserPreferences.PhysicsRate._45:  return "45Hz";
				case UserPreferences.PhysicsRate._60:  return "60Hz";
				case UserPreferences.PhysicsRate._72:  return "72Hz";
				case UserPreferences.PhysicsRate._80:  return "80Hz";
				case UserPreferences.PhysicsRate._90:  return "90Hz";
				case UserPreferences.PhysicsRate._120: return "120Hz";
				case UserPreferences.PhysicsRate._144: return "144Hz";
				case UserPreferences.PhysicsRate._240: return "240Hz";
				case UserPreferences.PhysicsRate._288: return "288Hz";
				default: return "Unknown";
			}
		}

		private static string MSAAToStr(int msaaLevel)
		{
			if (msaaLevel > 0)
				return msaaLevel + "x";
			else
				return "Off";
		}

		private static string Compare(string label, bool officialSetting, bool usedSetting)
		{
			if (officialSetting != usedSetting)
				return "<color=red>" + label + ": " + (usedSetting ? "On" : "Off") + "</color>";
			else
				return label + ": " + (usedSetting ? "On" : "Off");
		}

		private static string Compare(string label, int officialSetting, int usedSetting, string unit)
		{
			if (officialSetting != usedSetting)
				return "<color=red>" + label + ": " + usedSetting + unit + "</color>";
			else
				return label + ": " + usedSetting + unit;
		}

		private static string Compare(string label, float officialSetting, float usedSetting, string unit)
		{
			if (officialSetting != usedSetting)
				return "<color=red>" + label + ": " + usedSetting + unit + "</color>";
			else
				return label + ": " + usedSetting + unit;
		}

		private static string Compare(string label, string officialSetting, string usedSetting)
		{
			if (officialSetting != usedSetting)
				return "<color=red>" + label + ": " + usedSetting + "</color>";
			else
				return label + ": " + usedSetting;
		}

		// Overwrite UI material+settings. This makes worldspace UI loaded via CustomUnityAsset
		// appear properly in the 3D world, rather then being rendered on top of everything.
		// Don't forget to destroy the defaultMaterial when your script is destroyed.
		public static void FixUICanvas(Canvas canvas, ref Material defaultMaterial)
		{
			canvas.sortingOrder = -1;
			UserPreferences up = UserPreferences.singleton;
			if (defaultMaterial == null)
				defaultMaterial = new Material(up.panelForUIMaterial.defaultMaterial);
			if (up.defaultUIShader != null)
				defaultMaterial.shader = up.defaultUIShader;
			Graphic[] graphics = canvas.GetComponentsInChildren<Graphic>(true);
			for (int i=0; i<graphics.Length; ++i)
				graphics[i].material = defaultMaterial;
		}

		private class Record
		{
			public struct MinMaxAvg
			{
				public float Min { get { return myMin; } }
				public float Max { get { return myMax; } }
				public float Avg { get { return (float)myAvg; } }

				internal float myMin;
				internal float myMax;
				internal double myAvg;
			}

			public string Scene;
			public int Frames;
			public bool Valid;

			public MinMaxAvg FPS;
			public MinMaxAvg TotalTime;
			public MinMaxAvg RenderTime;
			public MinMaxAvg ScriptsTime;
			public MinMaxAvg PhysicsTime;
			public MinMaxAvg WaitTime;

			public void Init(string scene)
			{
				Scene = scene;
				Reset();
			}

			public void Reset()
			{
				Frames = 0;
				Valid = false;

				Reset(ref FPS);
				Reset(ref TotalTime);
				Reset(ref RenderTime);
				Reset(ref ScriptsTime);
				Reset(ref PhysicsTime);
				Reset(ref WaitTime);
			}

			public void Sample()
			{
				Sample(ref FPS, 1.0f / Time.deltaTime);
				Sample(ref TotalTime, PerfMon.totalTime);
				Sample(ref RenderTime, PerfMon.renderTime);
				Sample(ref ScriptsTime, PerfMon.scriptsTime);
				Sample(ref PhysicsTime, PerfMon.physicsTime);
				Sample(ref WaitTime, PerfMon.waitTime);
				++Frames;
			}

			public void Apply()
			{
				if (Frames > 0)
				{
					Apply(ref FPS, Frames);
					Apply(ref TotalTime, Frames);
					Apply(ref RenderTime, Frames);
					Apply(ref ScriptsTime, Frames);
					Apply(ref PhysicsTime, Frames);
					Apply(ref WaitTime, Frames);
					Valid = true;
				}
				else
				{
					Valid = false;
				}
			}

			private void Sample(ref MinMaxAvg data, float v)
			{
				data.myMin = Mathf.Min(data.myMin, v);
				data.myMax = Mathf.Max(data.myMax, v);
				data.myAvg += v;
			}

			private void Apply(ref MinMaxAvg data, int frames)
			{
				data.myAvg /= frames;
			}

			private void Reset(ref MinMaxAvg data)
			{
				data.myMin = float.MaxValue;
				data.myMax = 0.0f;
				data.myAvg = 0.0;
			}
		}

		private class FullRecord
		{
			private const int BUFFER_SIZE = 300*75; // Assuming a max of an average 300fps for a 75 second scene
			private const float QUANTILE = 0.01f;

			public List<float> FPS = new List<float>(BUFFER_SIZE);
			public List<float> TotalTime = new List<float>(BUFFER_SIZE);
			public List<float> RenderTime = new List<float>(BUFFER_SIZE);
			public List<float> ScriptsTime = new List<float>(BUFFER_SIZE);
			public List<float> PhysicsTime = new List<float>(BUFFER_SIZE);
			public List<float> WaitTime = new List<float>(BUFFER_SIZE);

			public string Scene;

			public void Init(string scene)
			{
				Scene = scene;
				Reset();
			}

			public void Reset()
			{
				FPS.Clear();
				TotalTime.Clear();
				RenderTime.Clear();
				ScriptsTime.Clear();
				PhysicsTime.Clear();
				WaitTime.Clear();
			}

			public void Sample()
			{
				FPS.Add(1.0f / Time.deltaTime);
				TotalTime.Add(PerfMon.totalTime);
				RenderTime.Add(PerfMon.renderTime);
				ScriptsTime.Add(PerfMon.scriptsTime);
				PhysicsTime.Add(PerfMon.physicsTime);
				WaitTime.Add(PerfMon.waitTime);
			}

			public Record ApplyAndGetRecord()
			{
				Record record = new Record();
				record.Init(Scene);

				Apply(ref FPS, ref record.FPS);
				Apply(ref TotalTime, ref record.TotalTime);
				Apply(ref RenderTime, ref record.RenderTime);
				Apply(ref ScriptsTime, ref record.ScriptsTime);
				Apply(ref PhysicsTime, ref record.PhysicsTime);
				Apply(ref WaitTime, ref record.WaitTime);

				return record;
			}

			private void Apply(ref List<float> frames, ref Record.MinMaxAvg result)
			{
				if (frames.Count == 0)
					return;

				result.myAvg = 0.0f;
				for (int i=0; i<frames.Count; ++i)
					result.myAvg += frames[i];
				result.myAvg /= frames.Count;

				frames.Sort();
				int minIdx = Mathf.Min(Mathf.CeilToInt(frames.Count * QUANTILE), frames.Count-1);
				int maxIdx = Mathf.FloorToInt(frames.Count * (1.0f-QUANTILE));
				result.myMin = frames[minIdx];
				result.myMax = frames[maxIdx];
			}
		}

		private class VaMSettings
		{
			public float renderScale;
			public float monitorCameraFOV;
			public bool mirrorReflections;
			public bool realtimeReflectionProbes;
			public int msaaLevel;
			public int pixelLightCount;
			public UserPreferences.ShaderLOD shaderLOD;
			public UserPreferences.GlowEffectsLevel glowEffects;
			public int smoothPasses;
			public UserPreferences.PhysicsRate physicsRate;
			public int physicsUpdateCap;
			public bool softPhysics;
			public bool physicsHighQuality;
			public bool desktopVsync;
			public bool mirrorToDisplay;
			public bool performanceMonitor;
			public float timeScale;
			public TimeControl.State timeState;

			public void Read()
			{
				SuperController sc = SuperController.singleton;
				UserPreferences up = UserPreferences.singleton;
				PerfMon pm = GameObject.Find("/SceneAtoms/CoreControl/PerfMon").GetComponent<PerfMon>();
				TimeControl tc = TimeControl.singleton;

				renderScale = up.renderScale;
				monitorCameraFOV = sc.monitorCameraFOV;
				mirrorReflections = up.mirrorReflections;
				realtimeReflectionProbes = up.realtimeReflectionProbes;
				msaaLevel = up.msaaLevel;
				pixelLightCount = up.pixelLightCount;
				shaderLOD = up.shaderLOD;
				glowEffects = up.glowEffects;
				smoothPasses = up.smoothPasses;
				physicsRate = up.physicsRate;
				physicsUpdateCap = up.physicsUpdateCap;
				softPhysics = up.softPhysics;
				physicsHighQuality = up.physicsHighQuality;
				desktopVsync = up.desktopVsync;
				mirrorToDisplay = up.mirrorToDisplay;
				performanceMonitor = pm.on;
				timeScale = tc.currentScale;
				timeState = tc.currentState;
			}

			public void Apply()
			{
				SuperController sc = SuperController.singleton;
				UserPreferences up = UserPreferences.singleton;
				PerfMon pm = GameObject.Find("/SceneAtoms/CoreControl/PerfMon").GetComponent<PerfMon>();
				TimeControl tc = TimeControl.singleton;

				up.renderScale = renderScale;
				sc.monitorCameraFOV = monitorCameraFOV;
				up.mirrorReflections = mirrorReflections;
				up.realtimeReflectionProbes = realtimeReflectionProbes;
				up.msaaLevel = msaaLevel;
				up.pixelLightCount = pixelLightCount;
				up.shaderLOD = shaderLOD;
				up.glowEffects = glowEffects;
				up.smoothPasses = smoothPasses;
				up.physicsRate = physicsRate;
				up.physicsUpdateCap = physicsUpdateCap;
				up.softPhysics = softPhysics;
				up.physicsHighQuality = physicsHighQuality;
				up.desktopVsync = desktopVsync;
				up.mirrorToDisplay = mirrorToDisplay;
				pm.on = performanceMonitor;
				tc.currentScale = timeScale;
				tc.currentState = timeState;
			}

			public static readonly VaMSettings Benchmark = new VaMSettings()
			{
				renderScale = 1.0f,
				monitorCameraFOV = 40.0f,
				mirrorReflections = true,
				realtimeReflectionProbes = false,
				msaaLevel = 8,
				pixelLightCount = 3,
				shaderLOD = UserPreferences.ShaderLOD.High,
				glowEffects = UserPreferences.GlowEffectsLevel.Off,
				smoothPasses = 4,
				physicsRate = UserPreferences.PhysicsRate._60,
				physicsUpdateCap = 3,
				softPhysics = true,
				physicsHighQuality = false,
				desktopVsync = false,
				mirrorToDisplay = false,
				performanceMonitor = false,
				timeScale = 1.0f,
				timeState = TimeControl.State.Normal
			};
		}
	}
}